//! Simplest sync [Monad].
//!
//! For async, see [`future`].
//!
//! For abstracting away dealing with fallible data, see [`option`]/[`result`].

#[cfg(doc)]
use crate::func::instances::*;
use crate::func::*;

#[derive(SharedFunctorAny)]
pub struct SoloInstance;

impl WeakFunctorAny for SoloInstance {
    type FAny<'a, A: 'a> = A;
}

impl<'a> Functor<'a> for SoloInstance {
    fn fmap<A: 'a, B: 'a>(fa: Self::F<A>, f: impl 'a + FnOnce(A) -> B) -> Self::F<B> {
        f(fa)
    }

    fn replace<A: 'a, B: 'a>(fa: Self::F<A>, b: B) -> Self::F<B> {
        drop(fa);
        b
    }

    fn void<A: 'a>(fa: Self::F<A>) -> Self::F<()> {
        drop(fa);
    }
}

impl<'a> Pure<'a> for SoloInstance {
    fn pure<A: 'a>(a: A) -> Self::F<A> {
        a
    }
}

impl<'a> ApplicativeSeq<'a> for SoloInstance {
    fn seq<A: 'a, B: 'a>(ff: Self::F<impl 'a + FnOnce(A) -> B>, fa: Self::F<A>) -> Self::F<B> {
        ff(fa)
    }
}

impl<'a> ApplicativeLA2<'a> for SoloInstance {
    fn la2<A: 'a, B: 'a, C: 'a>(
        fa: Self::F<A>,
        fb: Self::F<B>,
        f: impl 'a + FnOnce(A, B) -> C,
    ) -> Self::F<C> {
        f(fa, fb)
    }
}

impl<'a> ApplicativeTuple<'a> for SoloInstance {
    fn tuple<A: 'a, B: 'a>((fa, fb): (Self::F<A>, Self::F<B>)) -> Self::F<(A, B)> {
        (fa, fb)
    }
}

impl<'a> ApplicativeSelect<'a> for SoloInstance {}

impl<'a> Applicative<'a> for SoloInstance {
    fn discard_first<A: 'a, B: 'a>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<B> {
        drop(fa);
        fb
    }

    fn discard_second<A: 'a, B: 'a>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<A> {
        drop(fb);
        fa
    }
}

impl<'a> Monad<'a> for SoloInstance {
    fn bind<A: 'a, B: 'a>(fa: Self::F<A>, f: impl 'a + FnOnce(A) -> Self::F<B>) -> Self::F<B> {
        f(fa)
    }

    fn iterate<B: 'a>(mut f: impl Iterative<'a, T = Self, B = B>) -> Self::F<B> {
        loop {
            match f.next() {
                ControlFlow::Continue(next_f) => f = next_f,
                ControlFlow::Break(b) => return b,
            }
        }
    }

    fn join<A: 'a>(ffa: Self::F<Self::F<A>>) -> Self::F<A> {
        ffa
    }
}

impl<'a> LocalFunctor<'a> for SoloInstance {
    fn unstuff<A: 'a, B: 'a>(state: Self::F<ControlFlow<B, A>>) -> ControlFlow<Self::F<B>, A> {
        state
    }

    fn stuff<A: 'a, T: Pure<'a>>(fa: Self::F<T::F<A>>) -> T::F<Self::F<A>> {
        fa
    }
}

impl<'a> Fail<'a, std::convert::Infallible> for SoloInstance {
    fn fail<A: 'a>(e: std::convert::Infallible) -> Self::F<A> {
        match e {}
    }
}
