//! Fallible version of [`future`].
//!
//! Preferred over building it out of [`composition`].
//!
//! To embed extra errors, see [`overload::EmbedFail`].

use std::{future::Future, pin::Pin};

use futures::{
    future::{try_select, Either, Shared},
    try_join, FutureExt,
};

#[cfg(doc)]
use crate::func::instances::*;
use crate::func::*;

pub struct TryFutureInstance<E>(E);

impl<E> WeakFunctorAny for TryFutureInstance<E> {
    type FAny<'a, A: 'a> = Pin<Box<dyn 'a + Future<Output = Result<A, E>>>> where Self: 'a;
}

impl<'a, E: 'a> Functor<'a> for TryFutureInstance<E> {
    fn fmap<A: 'a, B: 'a>(fa: Self::F<A>, f: impl 'a + FnOnce(A) -> B) -> Self::F<B> {
        Box::pin(async { Ok(f(fa.await?)) })
    }

    fn replace<A: 'a, B: 'a>(fa: Self::F<A>, b: B) -> Self::F<B> {
        Box::pin(async {
            fa.await?;
            Ok(b)
        })
    }
}

impl<'a, E: 'a> Pure<'a> for TryFutureInstance<E> {
    fn pure<A: 'a>(a: A) -> Self::F<A> {
        Box::pin(async { Ok(a) })
    }
}

impl<'a, E: 'a> ApplicativeSeq<'a> for TryFutureInstance<E> {
    fn seq<A: 'a, B: 'a>(ff: Self::F<impl 'a + FnOnce(A) -> B>, fa: Self::F<A>) -> Self::F<B> {
        Box::pin(async {
            let (f, a) = try_join!(ff, fa)?;
            Ok(f(a))
        })
    }
}

impl<'a, E: 'a> ApplicativeLA2<'a> for TryFutureInstance<E> {
    fn la2<A: 'a, B: 'a, C: 'a>(
        fa: Self::F<A>,
        fb: Self::F<B>,
        f: impl 'a + FnOnce(A, B) -> C,
    ) -> Self::F<C> {
        Box::pin(async {
            let (a, b) = try_join!(fa, fb)?;
            Ok(f(a, b))
        })
    }
}

impl<'a, E: 'a> ApplicativeTuple<'a> for TryFutureInstance<E> {
    fn tuple<A: 'a, B: 'a>((fa, fb): (Self::F<A>, Self::F<B>)) -> Self::F<(A, B)> {
        Box::pin(async { try_join!(fa, fb) })
    }
}

impl<'a, E: 'a> ApplicativeSelect<'a> for TryFutureInstance<E> {
    fn select<A: 'a, B: 'a>(fa: Self::F<A>, fb: Self::F<B>) -> SelectedWrapped<'a, A, B, Self> {
        Box::pin(async {
            match try_select(fa, fb).await {
                Ok(Either::Left((a, fb))) => Ok(Selected::A(a, fb)),
                Ok(Either::Right((b, fa))) => Ok(Selected::B(fa, b)),
                Err(Either::Left((e, _))) => Err(e),
                Err(Either::Right((e, _))) => Err(e),
            }
        })
    }
}

impl<'a, E: 'a> Applicative<'a> for TryFutureInstance<E> {
    fn discard_first<A: 'a, B: 'a>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<B> {
        Box::pin(async { Ok(try_join!(fa, fb)?.1) })
    }

    fn discard_second<A: 'a, B: 'a>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<A> {
        Box::pin(async { Ok(try_join!(fa, fb)?.0) })
    }
}

impl<'a, E: 'a> Monad<'a> for TryFutureInstance<E> {
    fn bind<A: 'a, B: 'a>(fa: Self::F<A>, f: impl 'a + FnOnce(A) -> Self::F<B>) -> Self::F<B> {
        Box::pin(async { f(fa.await?).await })
    }

    fn iterate<B: 'a>(mut f: impl Iterative<'a, T = Self, B = B>) -> Self::F<B> {
        Box::pin(async move {
            loop {
                match f.next().await? {
                    ControlFlow::Continue(next_f) => f = next_f,
                    ControlFlow::Break(b) => return Ok(b),
                }
            }
        })
    }

    fn join<A: 'a>(ffa: Self::F<Self::F<A>>) -> Self::F<A> {
        Box::pin(async { ffa.await?.await })
    }
}

impl<'a, E: 'a + Clone> SharedFunctor<'a> for TryFutureInstance<E> {
    type Shared<A: 'a + Clone> = Shared<Pin<Box<dyn 'a + Future<Output = Result<A, E>>>>>;

    fn share<A: 'a + Clone>(fa: Self::F<A>) -> Self::Shared<A> {
        fa.shared()
    }

    fn unshare<A: 'a + Clone>(sa: Self::Shared<A>) -> Self::F<A> {
        Box::pin(sa)
    }
}

impl<'a, E: 'a> Fail<'a, E> for TryFutureInstance<E> {
    fn fail<A: 'a>(e: E) -> Self::F<A> {
        Box::pin(async { Err(e) })
    }
}

pub struct FutureFailAny;

impl<'a> MonadFailAny<'a> for FutureFailAny {
    type W<E: 'a> = TryFutureInstance<E>;

    type T = instances::future::FutureInstance;

    fn unstuff<A: 'a, E: 'a>(wa: WrapE<'a, A, E, Self>) -> Wrap<'a, Result<A, E>, Self::T> {
        wa
    }

    fn stuff<A: 'a, E: 'a>(fa: Wrap<'a, Result<A, E>, Self::T>) -> WrapE<'a, A, E, Self> {
        fa
    }

    fn map_err<A: 'a, E0: 'a, E1: 'a>(
        wa: WrapE<'a, A, E0, Self>,
        f: impl 'a + FnOnce(E0) -> E1,
    ) -> WrapE<'a, A, E1, Self> {
        Box::pin(async { wa.await.map_err(f) })
    }

    fn bind_err<A: 'a, E0: 'a, E1: 'a>(
        wa: WrapE<'a, A, E0, Self>,
        f: impl 'a + FnOnce(E0) -> WrapE<'a, A, E1, Self>,
    ) -> WrapE<'a, A, E1, Self> {
        Box::pin(async {
            match wa.await {
                Ok(a) => Ok(a),
                Err(e) => f(e).await,
            }
        })
    }

    fn bind<A: 'a, B: 'a, E0: 'a, E1: 'a>(
        wa: WrapE<'a, A, E0, Self>,
        f: impl 'a + FnOnce(Result<A, E0>) -> WrapE<'a, B, E1, Self>,
    ) -> WrapE<'a, B, E1, Self> {
        Box::pin(async { f(wa.await).await })
    }

    fn rotate_out<A: 'a, E0: 'a, E1: 'a>(
        wa: WrapE<'a, Result<A, E1>, E0, Self>,
    ) -> WrapE<'a, A, Result<E1, E0>, Self> {
        Box::pin(async {
            match wa.await {
                Ok(Ok(a)) => Ok(a),
                Ok(Err(e)) => Err(Ok(e)),
                Err(e) => Err(Err(e)),
            }
        })
    }
}
